@@ -18,7 +18,6 @@
 #include <config.h>
 #include "internal.h"
 
-#include <unistd.h>
 #include <json.h>
 #include "b64/cencode.h"
 #include "b64/cdecode.h"
@@ -26,12 +25,13 @@
 
 static int
 prepare_response2 (const char *encstr, const char *bdstr, const char *input,
-		   char **response)
+		   char **response, size_t * response_len)
 {
   int rc = U2FH_JSON_ERROR;
   struct json_object *jo = NULL, *enc = NULL, *bd = NULL, *key = NULL;
   char keyb64[256];
   size_t keylen = sizeof (keyb64);
+  const char *reply;
 
   enc = json_object_new_string (encstr);
   if (enc == NULL)
@@ -59,7 +59,22 @@ prepare_response2 (const char *encstr, const char *bdstr, const char *input,
   json_object_object_add (jo, "clientData", bd);
   json_object_object_add (jo, "keyHandle", key);
 
-  *response = strdup (json_object_to_json_string (jo));
+  reply = json_object_to_json_string (jo);
+  if (*response == NULL)
+    {
+      *response = strdup (reply);
+    }
+  else
+    {
+      if (strlen (reply) >= *response_len)
+	{
+	  rc = U2FH_SIZE_ERROR;
+	  *response_len = strlen (reply) + 1;
+	  goto done;
+	}
+      strcpy (*response, reply);
+    }
+  *response_len = strlen (reply);
   if (*response == NULL)
     rc = U2FH_MEMORY_ERROR;
   else
@@ -67,16 +82,19 @@ prepare_response2 (const char *encstr, const char *bdstr, const char *input,
 
 done:
   json_object_put (jo);
-  json_object_put (enc);
-  json_object_put (bd);
-  json_object_put (key);
+  if (!jo)
+    {
+      json_object_put (enc);
+      json_object_put (bd);
+      json_object_put (key);
+    }
 
   return rc;
 }
 
 static int
 prepare_response (const unsigned char *buf, int len, const char *bd,
-		  const char *input, char **response)
+		  const char *input, char **response, size_t * response_len)
 {
   base64_encodestate b64ctx;
   char b64enc[2048];
@@ -96,7 +114,7 @@ prepare_response (const unsigned char *buf, int len, const char *bd,
   cnt = base64_encode_block (bd, strlen (bd), bdstr, &b64ctx);
   base64_encode_blockend (bdstr + cnt, &b64ctx);
 
-  return prepare_response2 (b64enc, bdstr, input, response);
+  return prepare_response2 (b64enc, bdstr, input, response, response_len);
 }
 
 #define CHALLBINLEN 32
@@ -104,23 +122,11 @@ prepare_response (const unsigned char *buf, int len, const char *bd,
 #define MAXKHLEN 128
 #define NOTSATISFIED "\x69\x85"
 
-/**
- * u2fh_authenticate:
- * @devs: a device handle, from u2fh_devs_init() and u2fh_devs_discover().
- * @challenge: string with JSON data containing the challenge.
- * @origin: U2F origin URL.
- * @response: pointer to output string with JSON data.
- * @flags: set of ORed #u2fh_cmdflags values.
- *
- * Perform the U2F Authenticate operation.
- *
- * Returns: On success %U2FH_OK (integer 0) is returned, and on errors
- * an #u2fh_rc error code.
- */
-u2fh_rc
-u2fh_authenticate (u2fh_devs * devs,
-		   const char *challenge,
-		   const char *origin, char **response, u2fh_cmdflags flags)
+static u2fh_rc
+_u2fh_authenticate (u2fh_devs * devs,
+		    const char *challenge,
+		    const char *origin, char **response,
+		    size_t * response_len, u2fh_cmdflags flags)
 {
   unsigned char data[CHALLBINLEN + HOSIZE + MAXKHLEN + 1];
   unsigned char buf[MAXDATASIZE];
@@ -134,12 +140,8 @@ u2fh_authenticate (u2fh_devs * devs,
   size_t kh64len = sizeof (khb64);
   base64_decodestate b64;
   size_t khlen;
-  int skip_devices[devs->num_devices];
-  int skipped = 0;
   int iterations = 0;
 
-  memset (skip_devices, 0, sizeof (skip_devices));
-
   rc = get_fixed_json_data (challenge, "challenge", chalb64, &challen);
   if (rc != U2FH_OK)
     return rc;
@@ -167,26 +169,20 @@ u2fh_authenticate (u2fh_devs * devs,
 
   do
     {
-      int i;
-      if (iterations++ > 15)
-	{
-	  return U2FH_TIMEOUT_ERROR;
-	}
-      for (i = 0; i < devs->num_devices; i++)
+      struct u2fdevice *dev;
+      for (dev = devs->first; dev != NULL; dev = dev->next)
 	{
 	  unsigned char tmp_buf[MAXDATASIZE];
-	  if (skip_devices[i] != 0)
+	  if (iterations == 0)
 	    {
-	      continue;
+	      dev->skipped = 0;
 	    }
-	  if (!devs->devs[i].is_alive)
+	  else if (dev->skipped != 0)
 	    {
-	      skipped++;
-	      skip_devices[i] = 1;
 	      continue;
 	    }
 	  len = MAXDATASIZE;
-	  rc = send_apdu (devs, i, U2F_AUTHENTICATE, data,
+	  rc = send_apdu (devs, dev->id, U2F_AUTHENTICATE, data,
 			  HOSIZE + CHALLBINLEN + khlen + 1,
 			  flags & U2FH_REQUEST_USER_PRESENCE ? 3 : 7, tmp_buf,
 			  &len);
@@ -201,11 +197,16 @@ u2fh_authenticate (u2fh_devs * devs,
 	    }
 	  else if (memcmp (tmp_buf, NOTSATISFIED, 2) != 0)
 	    {
-	      skipped++;
-	      skip_devices[i] = 2;
-	      continue;
+	      dev->skipped = 1;
 	    }
-	  memcpy (buf, tmp_buf, len);
+	  else
+	    {
+	      memcpy (buf, tmp_buf, len);
+	    }
+	}
+      if (iterations++ > 15)
+	{
+	  return U2FH_TIMEOUT_ERROR;
 	}
       if (len == 2 && memcmp (buf, NOTSATISFIED, 2) == 0)
 	{
@@ -219,10 +220,65 @@ u2fh_authenticate (u2fh_devs * devs,
     {
       return U2FH_AUTHENTICATOR_ERROR;
     }
+  else if ((flags & U2FH_REQUEST_USER_PRESENCE) == 0
+	   && len == 2 && memcmp (buf, NOTSATISFIED, 2) == 0)
+    {
+      return U2FH_OK;
+    }
   if (len != 2)
     {
-      prepare_response (buf, len - 2, bd, challenge, response);
+      prepare_response (buf, len - 2, bd, challenge, response, response_len);
+      return U2FH_OK;
     }
 
-  return U2FH_OK;
+  return U2FH_TRANSPORT_ERROR;
+}
+
+/**
+ * u2fh_authenticate2:
+ * @devs: a device handle, from u2fh_devs_init() and u2fh_devs_discover().
+ * @challenge: string with JSON data containing the challenge.
+ * @origin: U2F origin URL.
+ * @response: pointer to string for output data
+ * @response_len: pointer to length of @response
+ * @flags: set of ORed #u2fh_cmdflags values.
+ *
+ * Perform the U2F Authenticate operation.
+ *
+ * Returns: On success %U2FH_OK (integer 0) is returned, and on errors
+ * an #u2fh_rc error code.
+ */
+u2fh_rc
+u2fh_authenticate2 (u2fh_devs * devs,
+		    const char *challenge,
+		    const char *origin, char *response, size_t * response_len,
+		    u2fh_cmdflags flags)
+{
+  return _u2fh_authenticate (devs, challenge, origin, &response, response_len,
+			     flags);
+}
+
+/**
+ * u2fh_authenticate:
+ * @devs: a device handle, from u2fh_devs_init() and u2fh_devs_discover().
+ * @challenge: string with JSON data containing the challenge.
+ * @origin: U2F origin URL.
+ * @response: pointer to pointer for output data
+ * @flags: set of ORed #u2fh_cmdflags values.
+ *
+ * Perform the U2F Authenticate operation.
+ *
+ * Returns: On success %U2FH_OK (integer 0) is returned, and on errors
+ * an #u2fh_rc error code.
+ */
+u2fh_rc
+u2fh_authenticate (u2fh_devs * devs,
+		   const char *challenge,
+		   const char *origin, char **response, u2fh_cmdflags flags)
+{
+  size_t response_len = 0;
+
+  *response = NULL;
+  return _u2fh_authenticate (devs, challenge, origin, response, &response_len,
+			     flags);
 }
